//@line 36 "/home/cltbld/build/tracemonkey-mobile-browser-maemo5-qt/tracemonkey/mobile/locales/en-US/mobile-l10n.js"

//@line 38 "/home/cltbld/build/tracemonkey-mobile-browser-maemo5-qt/tracemonkey/mobile/locales/en-US/mobile-l10n.js"

pref("general.useragent.locale", "en-US");
